/* 
 * ChatScreen.java 
 * 
 * Main screen for BerryCraft - spawn an MCIO thread to connect to the minecraft server and receive packets. 
 * Then, continually check game chat and status buffer, and update title/screen accordingly.  
 * Send any typed commands to the Minecraft game server as a chat message.
 */

package com.synthdreams;

import net.rim.device.api.system.Characters;
import net.rim.device.api.system.Display;
import net.rim.device.api.ui.Field;
import net.rim.device.api.ui.MenuItem;
import net.rim.device.api.ui.UiApplication;
import net.rim.device.api.ui.XYEdges;
import net.rim.device.api.ui.component.*;
import net.rim.device.api.ui.container.*;
import net.rim.device.api.ui.decor.Border;
import net.rim.device.api.ui.decor.BorderFactory;

class ChatScreen extends MainScreen {
   private MCIO _mcio;
   private Thread _mcioThread;
   private Settings _settings;

   // Field Managers
   private HorizontalFieldManager _outputManager;
   private HorizontalFieldManager _inputManager;
   private HorizontalFieldManager _topManager;
   
   // Fields
   private LabelField _titleField;
   private EditField _outputField;
   private EditField _inputField;
   
   // Menu Items
   private MenuItem _connectItem;
   private MenuItem _disconnectItem;
   private MenuItem _macro1Item;
   private MenuItem _macro2Item;
   private MenuItem _macro3Item;
   private MenuItem _settingsItem;
   
   
   public ChatScreen()
   {
      super(NO_VERTICAL_SCROLL);

      Border tempBorder;
      
      _mcio = new MCIO();
      _settings = new Settings();
      
      // Setup Field Managers
      _inputManager = new HorizontalFieldManager(HorizontalFieldManager.HORIZONTAL_SCROLL | HorizontalFieldManager.NO_HORIZONTAL_SCROLLBAR) {
         protected void sublayout(int width, int height) {
            super.sublayout((int)(Display.getWidth() * 1), height);
            setExtent((int)(Display.getWidth() * 1), getFont().getHeight() * 2 - 5);
         }                  
      };

      _outputManager = new HorizontalFieldManager(HorizontalFieldManager.VERTICAL_SCROLL | HorizontalFieldManager.VERTICAL_SCROLLBAR) {
         protected void sublayout(int width, int height) {
            super.sublayout((int)(Display.getWidth() * 1), height);
            setExtent((int)(Display.getWidth() * 1), Display.getHeight() - getFont().getHeight() * 2 - 5);
         }         
      };

      _topManager = new HorizontalFieldManager(HorizontalFieldManager.NO_HORIZONTAL_SCROLL | HorizontalFieldManager.NO_VERTICAL_SCROLL) {
         protected void sublayout(int width, int height) {
            super.sublayout(Display.getWidth(), Display.getHeight() - getFont().getHeight() * 2 - 5);
            setExtent(Display.getWidth(), Display.getHeight() - getFont().getHeight() * 2 - 5);
         }
      };      
      
      // Fields
      _titleField = new LabelField();
      _outputField = new EditField(Field.USE_ALL_WIDTH | Field.READONLY);      
      _inputField = new EditField(Field.USE_ALL_WIDTH) {
         protected boolean keyChar(char character, int status, int time) {
            if (character == Characters.ENTER) {
               if (_mcio.getConnected() && (_mcio.getError() == false)) {
                  _mcio.sendChatMessage(getText());
               }
               setText("");

               return true;
            }
            return super.keyChar(character, status, time);
         }
      };
      
      tempBorder = BorderFactory.createBevelBorder(new XYEdges(3, 0, 0, 0));
      _inputField.setBorder(tempBorder);
      _outputManager.add(_outputField);
      _inputManager.add(_inputField);
      
      // Menu Items
      _connectItem = new MenuItem("Connect", 10, 0) {
         public void run() {
            // Connect to server
            if (_mcioThread == null) {
               _mcio.setParams(_settings.getServer(), _settings.getName(), _settings.getPassword(), _settings.getLauncherVer(), _settings.getProtocolVer());
               _mcio.resetError();
               
               _mcioThread = new Thread(_mcio);
               _mcioThread.start();
            }
         }
      };
      
      _disconnectItem = new MenuItem("Disconnect", 10, 0) {
         public void run() {
            
            if (_mcio.getConnected() && (_mcio.getError() == false)) {
               _mcio.sendDisconnectKick("Normal Exit");
               _mcio.disconnect();
               _mcioThread = null;
            }
         }
      };
      
      _macro1Item = new MenuItem("Send Macro 1", 20, 0) {
         public void run() {
            if (_mcio.getConnected() && (_mcio.getError() == false)) {
               _mcio.sendChatMessage(_settings.getMacro1());
            }
         }
      };

      _macro2Item = new MenuItem("Send Macro 2", 30, 0) {
         public void run() {
            if (_mcio.getConnected() && (_mcio.getError() == false)) {
               _mcio.sendChatMessage(_settings.getMacro2());
            }
         }
      };
      
      _macro3Item = new MenuItem("Send Macro 3", 40, 0) {
         public void run() {
            if (_mcio.getConnected() && (_mcio.getError() == false)) {
               _mcio.sendChatMessage(_settings.getMacro3());
            }
         }
      };
           
      _settingsItem = new MenuItem("Settings", 50, 0) {
         public void run() {
            getUiEngine().pushScreen(_settings);
         }
      };     
      
      // Load Saved Settings
      _settings.loadSettings();
      
      // Set Title
      setTitle(_titleField);
      
      // Set Menu Items
      addMenuItem(_connectItem);
      addMenuItem(_disconnectItem);
      addMenuItem(_macro1Item);
      addMenuItem(_macro2Item);
      addMenuItem(_macro3Item);
      addMenuItem(_settingsItem);
      
      // Set Managers and Fields
      _topManager.add(_outputManager);
      add(_topManager);
      add(_inputManager);
      
      _inputField.setFocus();
      
      // Setup updating every 100ms
      UiApplication.getUiApplication().invokeLater(
            new Runnable() {
               public void run() {
                  String tempOutput = "";
                  int cursorPosition;
                  
                  for (int lcv = 0 ; lcv < Game.getChat().size(); lcv++) {
                     tempOutput += (String)Game.getChat().elementAt(lcv);
                  }
                  
                  // Maintain cursor position if user is scrolling back through chat buffer
                  cursorPosition = _outputField.getCursorPosition();
                  _outputField.setText(tempOutput);
                  
                  if (_outputField.isFocus()) {
                     _outputField.setCursorPosition(cursorPosition);
                  }
                  else {
                     _outputField.setFocus();
                     _outputField.setCursorPosition(_outputField.getText().length());
                     _inputField.setFocus();
                  }
                  
                  // Update title, display error if present, or normally displaye game time when connected
                  if (Game.getStatus() != "") {
                     _titleField.setText(Game.getStatus());
                  }
                  else {
                     if (_mcio.getConnected()) {
                        _titleField.setText("Connected | Game Time: " + (Game.getTime() % 24000));
                     }
                     else {
                        _titleField.setText("Disconnected.");
                     }
                  }          
                  
                  // Check for an error state in MCIO - if so, kill thread so we can restart it if desired
                  if (_mcioThread != null) {
                     if (_mcio.getError()) _mcioThread = null;
                  }
               } 
            }, 100, true);
   }
   
   
   // Disconnect from Minecraft on program exit
   public void close() {
      _mcio.disconnect();
      super.close();
   }
}
