/* 
 * MCIO.java 
 * 
 * Minecraft IO class implements the Minecraft protocol as documented by http://mc.kev009.com/Protocol
 * 
 * When MCIO thread initially runs, it will connect to the Minecraft game server, authenticate with 
 * Minecraft.net (if server is online mode), login, and then go into a loop receiving minecraft packets.
 * 
 * After receiving a packet, if there is a "process" function defined for it, it will execute it (usually pushing
 * information into the Game class to be used by the Chat Screen)
 * 
 * Other classes (such as the Chat Screen) may send commands using a "send" function.
 * 
 * See "run" method for setup of receive loop
 * See "connect", "receiveCommand" and "sendCommand" for primary I/O methods. 
 * See "requestSession" and "requestSession" for Minecraft.net related functions
 * 
 * If you use this library in your program, please make note and link to http://www.ToniWestbrook.com
 * and keep reference to http://mc.kev009.com/Protocol
 * 
 */

package com.synthdreams;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.util.Vector;

import javax.microedition.io.HttpConnection;
import javax.microedition.io.HttpsConnection;
import javax.microedition.io.StreamConnection;

import net.rim.blackberry.api.browser.URLEncodedPostData;
import net.rim.device.api.io.transport.*;

public class MCIO implements Runnable {
   public static final int ARGUMENTS_MAX = 10;
   
   // Packet parameter datatypes
   public static final int TYPE_NONE = 0;
   public static final int TYPE_BOOLEAN = 1;
   public static final int TYPE_BYTE = 2;
   public static final int TYPE_SHORT = 3;
   public static final int TYPE_INT = 4;
   public static final int TYPE_LONG = 5;
   public static final int TYPE_FLOAT = 6;
   public static final int TYPE_DOUBLE = 7;
   public static final int TYPE_STRING = 8;
   public static final int TYPE_SPECIAL = 9;
 
   // Setup byte code for each Minecraft packet type
   public static final byte ID_KEEP_ALIVE = 0x00;
   public static final byte ID_LOGIN_REQUEST = 0x01;
   public static final byte ID_HANDSHAKE = 0x02;
   public static final byte ID_CHAT_MESSAGE = 0x03;
   public static final byte ID_TIME_UPDATE = 0x04;
   public static final byte ID_ENTITY_EQUIPMENT = 0x05;
   public static final byte ID_SPAWN_POSITION = 0x06;
   public static final byte ID_USE_ENTITY = 0x07;
   public static final byte ID_UPDATE_HEALTH = 0x08;
   public static final byte ID_RESPAWN = 0x09;
   public static final byte ID_PLAYER = 0x0A;
   public static final byte ID_PLAYER_POSITION= 0x0B;
   public static final byte ID_PLAYER_LOOK = 0x0C;
   public static final byte ID_PLAYER_POSITION_LOOK = 0x0D;
   public static final byte ID_PLAYER_DIGGING = 0x0E;
   public static final byte ID_PLAYER_BLOCK_PLACEMENT = 0x0F;
   public static final byte ID_HELD_ITEM_CHANGED = 0x10;
   public static final byte ID_USE_BED = 0x11;
   public static final byte ID_ANIMATION = 0x12;
   public static final byte ID_ENTITY_ACTION = 0x13;
   public static final byte ID_SPAWN_NAMED_ENTITY = 0x14;
   public static final byte ID_SPAWN_DROPPED_ITEM = 0x15;
   public static final byte ID_COLLECT_ITEM = 0x16;
   public static final byte ID_SPAWN_OBJECT_VEHICLE = 0x17;
   public static final byte ID_SPAWN_MOB = 0x18;
   public static final byte ID_SPAWN_PAINTING = 0x19;
   public static final byte ID_SPAWN_EXPERIENCE_ORB = 0x1A;
   public static final byte ID_ENTITY_VELOCITY = 0x1C;
   public static final byte ID_DESTROY_ENTITY = 0x1D;
   public static final byte ID_ENTITY = 0x1E;
   public static final byte ID_ENTITY_RELATIVE_MOVE = 0x1F;
   public static final byte ID_ENTITY_LOOK = 0x20;
   public static final byte ID_ENTITY_LOOK_RELATIVE_MOVE = 0x21;
   public static final byte ID_ENTITY_TELEPORT = 0x22;
   public static final byte ID_ENTITY_HEAD_LOOK = 0x23;
   public static final byte ID_ENTITY_STATUS = 0x26;
   public static final byte ID_ATTACH_ENTITY = 0x27;
   public static final byte ID_ENTITY_METADATA = 0x28;
   public static final byte ID_ENTITY_EFFECT = 0x29;
   public static final byte ID_REMOVE_ENTITY_EFFECT = 0x2A;
   public static final byte ID_SET_EXPERIENCE = 0x2B;
   public static final byte ID_MAP_COLUMN_ALLOCATION = 0x32;
   public static final byte ID_MAP_CHUNKS = 0x33;
   public static final byte ID_MULTI_BLOCK_CHANGE = 0x34;
   public static final byte ID_BLOCK_CHANGE = 0x35;
   public static final byte ID_BLOCK_ACTION = 0x36;
   public static final byte ID_EXPLOSION = 0x3C;
   public static final byte ID_SOUND_PARTICLE_EFFECT = 0x3D;
   public static final byte ID_CHANGE_GAME_STATE = 0x46;
   public static final byte ID_THUNDERBOLT = 0x47;
   public static final byte ID_OPEN_WINDOW = 0x64;
   public static final byte ID_CLOSE_WINDOW = 0x65;
   public static final byte ID_CLICK_WINDOW = 0x66;
   public static final byte ID_SET_SLOT = 0x67;
   public static final byte ID_SET_WINDOW_ITEMS = 0x68;
   public static final byte ID_UPDATE_WINDOW_PROPERTY = 0x69;
   public static final byte ID_CONFIRM_TRANSACTION = 0x6A;
   public static final byte ID_CREATIVE_INVENTORY_ACTION = 0x6B;
   public static final byte ID_ENCHANT_ITEM = 0x6C;
   public static final byte ID_UPDATE_SIGN = (byte) 0x82;
   public static final byte ID_ITEM_DATA = (byte) 0x83;
   public static final byte ID_UPDATE_TILE_ENTITY = (byte) 0x84;
   public static final byte ID_INCREMENT_STATISTIC = (byte) 0xC8;
   public static final byte ID_PLAYER_LIST_ITEM = (byte) 0xC9;
   public static final byte ID_PLAYER_ABILITIES = (byte) 0xCA;
   public static final byte ID_PLUGIN_MESSAGE = (byte) 0xFA;
   public static final byte ID_SERVER_LIST_PING = (byte) 0xFE;
   public static final byte ID_DISCONNECT_KICK = (byte) 0xFF;
   
   // Error Codes
   public static final int RETURN_OK = 0;
   public static final int RETURN_BADCONNECT = 1;
   public static final int RETURN_BADDISCONNECT = 2;
   public static final int RETURN_BADAUTH = 3;
   public static final int RETURN_BADJOIN = 4;
   public static final int RETURN_PACKETERROR = 5;
   
   // Send and Receive Argument Arrays
   private PacketArgument[][] _sendArgs;
   private PacketArgument[][] _recvArgs;
   
   // Streamer I/O for Socket
   private StreamConnection _serverStream;
   private DataOutputStream _serverWriter;
   private DataInputStream _serverReader;
   
   // Connection and authentication info
   private String _serverName, _userName, _password, _launcherVer, _protocolVer;
   
   // Session ID and server hash gotten from Minecraft.net servers
   private String _sessionID;
   private String _serverHash;
   
   // Connected and error state
   private boolean _connected;
   private boolean _error;
   
   // Return unsigned byte for deferencing an array
   private static int getByteIndex(byte passIndex) {
      if (passIndex < 0 ) return (int)passIndex + 256;
      else return (int)passIndex;
   }

   // Convert error code to message
   public static String getErrorMessage(int passError) {
      switch (passError) {
         case RETURN_OK: return "OK";
         case RETURN_BADCONNECT: return "Unable to connect to game server.";  
         case RETURN_BADDISCONNECT: return "Unable to disconnect.";
         case RETURN_BADAUTH: return "Bad username or password.";
         case RETURN_BADJOIN: return "Unable to join Minecraft.net.";
         case RETURN_PACKETERROR: return "Packet error.";
         default: return "Unknown Error.";
      }
   }   
   
   // Return if item is enchantable
   private static boolean enchantFlag(short passValue) {
      return (256 <= passValue && passValue <= 259) ||
             (267 <= passValue && passValue <= 279) ||
             (283 <= passValue && passValue <= 286) ||
             (290 <= passValue && passValue <= 294) ||
             (298 <= passValue && passValue <= 317) ||
             passValue == 261 || passValue == 359 || passValue == 346;
   }
     
   // Make an HTTP GET request - URL should include querystring
   private static String httpRequest(String passURL) throws IOException {
      ConnectionFactory tempFactory;
      ConnectionDescriptor tempDescriptor;
      HttpConnection mcConnection;
      byte[] mcResponseBytes;
      InputStream inStream;

      mcResponseBytes = new byte[8192];      
      
      // Setup Connection
      tempFactory = new ConnectionFactory();
      tempDescriptor = tempFactory.getConnection(passURL);
      
      mcConnection = (HttpConnection) tempDescriptor.getConnection();
      mcConnection.setRequestProperty("User-Agent", "Java/Blackberry");
      
      // Send Request
      mcConnection.getResponseCode();
         
      // Get Response
      inStream = mcConnection.openInputStream();
      inStream.read(mcResponseBytes);
      return new String(mcResponseBytes);      
   }  
   
   // Make an HTTPS POST request
   private static String httpsRequest(String passURL, URLEncodedPostData passParams) throws IOException {
      ConnectionFactory tempFactory;
      ConnectionDescriptor tempDescriptor;
      HttpsConnection mcConnection;
      byte[] mcPostBytes;
      byte[] mcResponseBytes;
      InputStream inStream;
      OutputStream outStream;

      mcResponseBytes = new byte[8192];      
      mcPostBytes = passParams.getBytes();
      
      // Setup Connection
      tempFactory = new ConnectionFactory();
      tempDescriptor = tempFactory.getConnection(passURL);
      mcConnection = (HttpsConnection) tempDescriptor.getConnection();
      mcConnection.setRequestMethod(HttpsConnection.POST);
      mcConnection.setRequestProperty("User-Agent", "Java/Blackberry");
      mcConnection.setRequestProperty("Content-Language", "en-US");
      mcConnection.setRequestProperty("Content-Type", "application/x-www-form-urlencoded");
      mcConnection.setRequestProperty("Content-Length", (new Integer(mcPostBytes.length)).toString());
      
      // Send Request
      outStream = mcConnection.openOutputStream();
      outStream.write(mcPostBytes);
      outStream.flush();
      mcConnection.getResponseCode();

      // Get Response
      inStream = mcConnection.openInputStream();
      inStream.read(mcResponseBytes);

      return new String(mcResponseBytes);      
   }
   
   public MCIO() {
      _serverName = "";
      _userName = "";
      _password = "";
      _launcherVer = "";
      _protocolVer = "";
      
      _sessionID = "";
      _serverHash = "";
      _connected = false;
      _error = false;
      _sendArgs = new PacketArgument[256][ARGUMENTS_MAX];
      _recvArgs = new PacketArgument[256][ARGUMENTS_MAX];
      
      for (int lcv = 0 ; lcv < 256 ; lcv++) {
         for (int lcv2 = 0 ; lcv2 < ARGUMENTS_MAX ; lcv2++) {
            _sendArgs[lcv][lcv2] = new PacketArgument();
            _recvArgs[lcv][lcv2] = new PacketArgument();
         }
      }
          
      // Define argument sizes and counts for each Minecraft packet type
      _sendArgs[getByteIndex(ID_KEEP_ALIVE)][0].argType = TYPE_INT; 
      _sendArgs[getByteIndex(ID_LOGIN_REQUEST)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_LOGIN_REQUEST)][1].argType = TYPE_STRING;
      _sendArgs[getByteIndex(ID_LOGIN_REQUEST)][2].argType = TYPE_STRING;
      _sendArgs[getByteIndex(ID_LOGIN_REQUEST)][3].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_LOGIN_REQUEST)][4].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_LOGIN_REQUEST)][5].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_LOGIN_REQUEST)][6].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_LOGIN_REQUEST)][7].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_HANDSHAKE)][0].argType = TYPE_STRING;
      _sendArgs[getByteIndex(ID_CHAT_MESSAGE)][0].argType = TYPE_STRING;
      _sendArgs[getByteIndex(ID_TIME_UPDATE)][0].argType = TYPE_LONG;
      _sendArgs[getByteIndex(ID_ENTITY_EQUIPMENT)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_ENTITY_EQUIPMENT)][1].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_ENTITY_EQUIPMENT)][2].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_ENTITY_EQUIPMENT)][3].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_SPAWN_POSITION)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_POSITION)][1].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_POSITION)][2].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_USE_ENTITY)][0].argType = TYPE_INT;      
      _sendArgs[getByteIndex(ID_USE_ENTITY)][1].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_USE_ENTITY)][2].argType = TYPE_BOOLEAN;
      _sendArgs[getByteIndex(ID_UPDATE_HEALTH)][0].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_UPDATE_HEALTH)][1].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_UPDATE_HEALTH)][2].argType = TYPE_FLOAT;
      _sendArgs[getByteIndex(ID_RESPAWN)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_RESPAWN)][1].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_RESPAWN)][2].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_RESPAWN)][3].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_RESPAWN)][4].argType = TYPE_STRING;
      _sendArgs[getByteIndex(ID_PLAYER)][0].argType = TYPE_BOOLEAN;
      _sendArgs[getByteIndex(ID_PLAYER_POSITION)][0].argType = TYPE_DOUBLE;
      _sendArgs[getByteIndex(ID_PLAYER_POSITION)][1].argType = TYPE_DOUBLE;
      _sendArgs[getByteIndex(ID_PLAYER_POSITION)][2].argType = TYPE_DOUBLE;
      _sendArgs[getByteIndex(ID_PLAYER_POSITION)][3].argType = TYPE_DOUBLE;
      _sendArgs[getByteIndex(ID_PLAYER_POSITION)][4].argType = TYPE_BOOLEAN;
      _sendArgs[getByteIndex(ID_PLAYER_LOOK)][0].argType = TYPE_FLOAT;
      _sendArgs[getByteIndex(ID_PLAYER_LOOK)][1].argType = TYPE_FLOAT;
      _sendArgs[getByteIndex(ID_PLAYER_LOOK)][2].argType = TYPE_BOOLEAN;
      _sendArgs[getByteIndex(ID_PLAYER_POSITION_LOOK)][0].argType = TYPE_DOUBLE;
      _sendArgs[getByteIndex(ID_PLAYER_POSITION_LOOK)][1].argType = TYPE_DOUBLE;
      _sendArgs[getByteIndex(ID_PLAYER_POSITION_LOOK)][2].argType = TYPE_DOUBLE;
      _sendArgs[getByteIndex(ID_PLAYER_POSITION_LOOK)][3].argType = TYPE_DOUBLE;
      _sendArgs[getByteIndex(ID_PLAYER_POSITION_LOOK)][4].argType = TYPE_FLOAT;
      _sendArgs[getByteIndex(ID_PLAYER_POSITION_LOOK)][5].argType = TYPE_FLOAT;
      _sendArgs[getByteIndex(ID_PLAYER_POSITION_LOOK)][6].argType = TYPE_BOOLEAN;
      _sendArgs[getByteIndex(ID_PLAYER_DIGGING)][0].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_PLAYER_DIGGING)][1].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_PLAYER_DIGGING)][2].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_PLAYER_DIGGING)][3].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_PLAYER_DIGGING)][4].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_PLAYER_BLOCK_PLACEMENT)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_PLAYER_BLOCK_PLACEMENT)][1].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_PLAYER_BLOCK_PLACEMENT)][2].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_PLAYER_BLOCK_PLACEMENT)][3].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_PLAYER_BLOCK_PLACEMENT)][4].argType = TYPE_SPECIAL;
      _sendArgs[getByteIndex(ID_HELD_ITEM_CHANGED)][0].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_USE_BED)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_USE_BED)][1].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_USE_BED)][2].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_USE_BED)][3].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_USE_BED)][4].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_ANIMATION)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_ANIMATION)][1].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_ENTITY_ACTION)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_ENTITY_ACTION)][1].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_SPAWN_NAMED_ENTITY)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_NAMED_ENTITY)][1].argType = TYPE_STRING;
      _sendArgs[getByteIndex(ID_SPAWN_NAMED_ENTITY)][2].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_NAMED_ENTITY)][3].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_NAMED_ENTITY)][4].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_NAMED_ENTITY)][5].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_SPAWN_NAMED_ENTITY)][6].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_SPAWN_NAMED_ENTITY)][7].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_SPAWN_DROPPED_ITEM)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_DROPPED_ITEM)][1].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_SPAWN_DROPPED_ITEM)][2].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_SPAWN_DROPPED_ITEM)][3].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_SPAWN_DROPPED_ITEM)][4].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_DROPPED_ITEM)][5].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_DROPPED_ITEM)][6].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_DROPPED_ITEM)][7].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_SPAWN_DROPPED_ITEM)][8].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_SPAWN_DROPPED_ITEM)][9].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_COLLECT_ITEM)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_COLLECT_ITEM)][1].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_OBJECT_VEHICLE)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_OBJECT_VEHICLE)][1].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_SPAWN_OBJECT_VEHICLE)][2].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_OBJECT_VEHICLE)][3].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_OBJECT_VEHICLE)][4].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_OBJECT_VEHICLE)][5].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_OBJECT_VEHICLE)][6].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_SPAWN_OBJECT_VEHICLE)][7].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_SPAWN_OBJECT_VEHICLE)][8].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_SPAWN_MOB)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_MOB)][1].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_SPAWN_MOB)][2].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_MOB)][3].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_MOB)][4].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_MOB)][5].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_SPAWN_MOB)][6].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_SPAWN_MOB)][7].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_SPAWN_MOB)][8].argType = TYPE_SPECIAL;
      _sendArgs[getByteIndex(ID_SPAWN_PAINTING)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_PAINTING)][1].argType = TYPE_STRING;
      _sendArgs[getByteIndex(ID_SPAWN_PAINTING)][2].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_PAINTING)][3].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_PAINTING)][4].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_PAINTING)][5].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_EXPERIENCE_ORB)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_EXPERIENCE_ORB)][1].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_EXPERIENCE_ORB)][2].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_EXPERIENCE_ORB)][3].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SPAWN_EXPERIENCE_ORB)][4].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_ENTITY_VELOCITY)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_ENTITY_VELOCITY)][1].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_ENTITY_VELOCITY)][2].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_ENTITY_VELOCITY)][3].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_DESTROY_ENTITY)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_ENTITY)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_ENTITY_RELATIVE_MOVE)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_ENTITY_RELATIVE_MOVE)][1].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_ENTITY_RELATIVE_MOVE)][2].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_ENTITY_RELATIVE_MOVE)][3].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_ENTITY_LOOK)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_ENTITY_LOOK)][1].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_ENTITY_LOOK)][2].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_ENTITY_LOOK_RELATIVE_MOVE)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_ENTITY_LOOK_RELATIVE_MOVE)][1].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_ENTITY_LOOK_RELATIVE_MOVE)][2].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_ENTITY_LOOK_RELATIVE_MOVE)][3].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_ENTITY_LOOK_RELATIVE_MOVE)][4].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_ENTITY_LOOK_RELATIVE_MOVE)][5].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_ENTITY_TELEPORT)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_ENTITY_TELEPORT)][1].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_ENTITY_TELEPORT)][2].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_ENTITY_TELEPORT)][3].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_ENTITY_TELEPORT)][4].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_ENTITY_TELEPORT)][5].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_ENTITY_HEAD_LOOK)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_ENTITY_HEAD_LOOK)][1].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_ENTITY_STATUS)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_ENTITY_STATUS)][1].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_ATTACH_ENTITY)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_ATTACH_ENTITY)][1].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_ENTITY_METADATA)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_ENTITY_METADATA)][1].argType = TYPE_SPECIAL;
      _sendArgs[getByteIndex(ID_ENTITY_EFFECT)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_ENTITY_EFFECT)][1].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_ENTITY_EFFECT)][2].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_ENTITY_EFFECT)][3].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_REMOVE_ENTITY_EFFECT)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_REMOVE_ENTITY_EFFECT)][1].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_SET_EXPERIENCE)][0].argType = TYPE_FLOAT;
      _sendArgs[getByteIndex(ID_SET_EXPERIENCE)][1].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_SET_EXPERIENCE)][2].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_MAP_COLUMN_ALLOCATION)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_MAP_COLUMN_ALLOCATION)][1].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_MAP_COLUMN_ALLOCATION)][2].argType = TYPE_BOOLEAN;
      _sendArgs[getByteIndex(ID_MAP_CHUNKS)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_MAP_CHUNKS)][1].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_MAP_CHUNKS)][2].argType = TYPE_BOOLEAN;
      _sendArgs[getByteIndex(ID_MAP_CHUNKS)][3].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_MAP_CHUNKS)][4].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_MAP_CHUNKS)][5].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_MAP_CHUNKS)][6].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_MAP_CHUNKS)][7].argType = TYPE_SPECIAL;
      _sendArgs[getByteIndex(ID_MULTI_BLOCK_CHANGE)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_MULTI_BLOCK_CHANGE)][1].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_MULTI_BLOCK_CHANGE)][2].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_MULTI_BLOCK_CHANGE)][3].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_MULTI_BLOCK_CHANGE)][4].argType = TYPE_SPECIAL;
      _sendArgs[getByteIndex(ID_BLOCK_CHANGE)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_BLOCK_CHANGE)][1].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_BLOCK_CHANGE)][2].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_BLOCK_CHANGE)][3].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_BLOCK_CHANGE)][4].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_BLOCK_ACTION)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_BLOCK_ACTION)][1].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_BLOCK_ACTION)][2].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_BLOCK_ACTION)][3].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_BLOCK_ACTION)][4].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_EXPLOSION)][0].argType = TYPE_DOUBLE;
      _sendArgs[getByteIndex(ID_EXPLOSION)][1].argType = TYPE_DOUBLE;
      _sendArgs[getByteIndex(ID_EXPLOSION)][2].argType = TYPE_DOUBLE;
      _sendArgs[getByteIndex(ID_EXPLOSION)][3].argType = TYPE_FLOAT;
      _sendArgs[getByteIndex(ID_EXPLOSION)][4].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_EXPLOSION)][5].argType = TYPE_SPECIAL;
      _sendArgs[getByteIndex(ID_SOUND_PARTICLE_EFFECT)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SOUND_PARTICLE_EFFECT)][1].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SOUND_PARTICLE_EFFECT)][2].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_SOUND_PARTICLE_EFFECT)][3].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_SOUND_PARTICLE_EFFECT)][4].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_CHANGE_GAME_STATE)][0].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_CHANGE_GAME_STATE)][1].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_THUNDERBOLT)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_THUNDERBOLT)][1].argType = TYPE_BOOLEAN;
      _sendArgs[getByteIndex(ID_THUNDERBOLT)][2].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_THUNDERBOLT)][3].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_THUNDERBOLT)][4].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_OPEN_WINDOW)][0].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_OPEN_WINDOW)][1].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_OPEN_WINDOW)][2].argType = TYPE_STRING;
      _sendArgs[getByteIndex(ID_OPEN_WINDOW)][3].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_CLOSE_WINDOW)][0].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_CLICK_WINDOW)][0].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_CLICK_WINDOW)][1].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_CLICK_WINDOW)][2].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_CLICK_WINDOW)][3].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_CLICK_WINDOW)][4].argType = TYPE_BOOLEAN;
      _sendArgs[getByteIndex(ID_CLICK_WINDOW)][5].argType = TYPE_SPECIAL;
      _sendArgs[getByteIndex(ID_SET_SLOT)][0].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_SET_SLOT)][1].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_SET_SLOT)][2].argType = TYPE_SPECIAL;
      _sendArgs[getByteIndex(ID_SET_WINDOW_ITEMS)][0].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_SET_WINDOW_ITEMS)][1].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_SET_WINDOW_ITEMS)][2].argType = TYPE_SPECIAL;
      _sendArgs[getByteIndex(ID_UPDATE_WINDOW_PROPERTY)][0].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_UPDATE_WINDOW_PROPERTY)][1].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_UPDATE_WINDOW_PROPERTY)][2].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_CONFIRM_TRANSACTION)][0].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_CONFIRM_TRANSACTION)][1].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_CONFIRM_TRANSACTION)][2].argType = TYPE_BOOLEAN;
      _sendArgs[getByteIndex(ID_CREATIVE_INVENTORY_ACTION)][0].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_CREATIVE_INVENTORY_ACTION)][1].argType = TYPE_SPECIAL;
      _sendArgs[getByteIndex(ID_ENCHANT_ITEM)][0].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_ENCHANT_ITEM)][1].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_UPDATE_SIGN)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_UPDATE_SIGN)][1].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_UPDATE_SIGN)][2].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_UPDATE_SIGN)][3].argType = TYPE_STRING;
      _sendArgs[getByteIndex(ID_UPDATE_SIGN)][4].argType = TYPE_STRING;
      _sendArgs[getByteIndex(ID_UPDATE_SIGN)][5].argType = TYPE_STRING;
      _sendArgs[getByteIndex(ID_UPDATE_SIGN)][6].argType = TYPE_STRING;
      _sendArgs[getByteIndex(ID_ITEM_DATA)][0].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_ITEM_DATA)][1].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_ITEM_DATA)][2].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_ITEM_DATA)][3].argType = TYPE_SPECIAL;
      _sendArgs[getByteIndex(ID_UPDATE_TILE_ENTITY)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_UPDATE_TILE_ENTITY)][1].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_UPDATE_TILE_ENTITY)][2].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_UPDATE_TILE_ENTITY)][3].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_UPDATE_TILE_ENTITY)][4].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_UPDATE_TILE_ENTITY)][5].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_UPDATE_TILE_ENTITY)][6].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_INCREMENT_STATISTIC)][0].argType = TYPE_INT;
      _sendArgs[getByteIndex(ID_INCREMENT_STATISTIC)][1].argType = TYPE_BYTE;
      _sendArgs[getByteIndex(ID_PLAYER_LIST_ITEM)][0].argType = TYPE_STRING;
      _sendArgs[getByteIndex(ID_PLAYER_LIST_ITEM)][1].argType = TYPE_BOOLEAN;
      _sendArgs[getByteIndex(ID_PLAYER_LIST_ITEM)][2].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_PLAYER_ABILITIES)][0].argType = TYPE_BOOLEAN;
      _sendArgs[getByteIndex(ID_PLAYER_ABILITIES)][1].argType = TYPE_BOOLEAN;
      _sendArgs[getByteIndex(ID_PLAYER_ABILITIES)][2].argType = TYPE_BOOLEAN;
      _sendArgs[getByteIndex(ID_PLAYER_ABILITIES)][3].argType = TYPE_BOOLEAN;
      _sendArgs[getByteIndex(ID_PLUGIN_MESSAGE)][0].argType = TYPE_STRING;
      _sendArgs[getByteIndex(ID_PLUGIN_MESSAGE)][1].argType = TYPE_SHORT;
      _sendArgs[getByteIndex(ID_PLUGIN_MESSAGE)][2].argType = TYPE_SPECIAL;
      _sendArgs[getByteIndex(ID_DISCONNECT_KICK)][0].argType = TYPE_STRING;

      // Copy send array to receive array
      for (int lcv = 0 ; lcv < 256 ; lcv++) {
         for (int lcv2 = 0 ; lcv2 < ARGUMENTS_MAX ; lcv2++) {
            _recvArgs[lcv][lcv2].argType = _sendArgs[lcv][lcv2].argType; 
         }
      }      
   }
   
   // Set server and authentication parameters
   public void setParams(String passServer, String passUser, String passPass, String passLauncher, String passProtocol) {
      _serverName = passServer;
      _userName = passUser;
      _password = passPass;      
      _launcherVer = passLauncher;
      _protocolVer = passProtocol;
   }
   
   // When the MCIO thread is spawned, it will initially attempt to login to the minecraft game server,
   // then go into a receive loop while connected, always processing any incoming packets.  When the server
   // disconnected, this function returns and the thread terminates.
   public void run() {
      int retVal;

      _error = false;
      Game.setStatus("Connecting...");
      
      retVal = connect();
      if (retVal != MCIO.RETURN_OK) { die(retVal); return;}
      
      Game.setStatus("Logging In...");
      
      // Send and receive handshake
      retVal = sendHandshake();
      if (retVal != MCIO.RETURN_OK) { die(retVal); return;}
      retVal = receiveCommand();
      if (retVal != MCIO.RETURN_OK) { die(retVal); return;}
      
      // Send login
      retVal = sendLoginRequest();
      if (retVal != MCIO.RETURN_OK) { die(retVal); return;}            
      retVal = receiveCommand();
      if (retVal != MCIO.RETURN_OK) { die(retVal); return;}
      
      Game.setStatus("");
      
      // Continually receive and process packets while connected
      while (getConnected() == true) {     
         retVal = receiveCommand();
         if (retVal != MCIO.RETURN_OK) {
            // If we had processing issues and are still connected, report error.  Otherwise it was a normal disconnect
            if (getConnected()) die(retVal);
            return;
         }
      }
   }
   
   // Request a session ID from Minecraft.net
   private boolean requestSession() {
      URLEncodedPostData mcPostData;
      String mcResponseString;
      
      mcPostData = new URLEncodedPostData(URLEncodedPostData.DEFAULT_CHARSET, false);
      mcPostData.append("user", _userName);
      mcPostData.append("password", _password);
      mcPostData.append("version", _launcherVer);
      
      try {
         mcResponseString = httpsRequest("https://login.minecraft.net", mcPostData);
      }
      catch (IOException e) {
         return false;
      }
      
      // Return false if bad login
      if (mcResponseString.lastIndexOf(':') == -1) return false;     
      
      // Set session ID
      _sessionID = mcResponseString.substring(mcResponseString.lastIndexOf(':')+1).trim();
      return true;
   }
   
   // Request to join the Minecraft.net servers
   private boolean requestJoin() {
      String joinURL;
      String mcResponseString;
      
      joinURL = "http://session.minecraft.net/game/joinserver.jsp";
      joinURL += "?user=" + _userName;
      joinURL += "&sessionId=" + _sessionID;
      joinURL += "&serverId=" + _serverHash;
      
      try {
         mcResponseString = httpRequest(joinURL);
      }
      catch (IOException e) {
         return false;
      }
            
      return (mcResponseString.startsWith("OK"));
   }
   
   // Return if connected to a game server
   public boolean getConnected() {
      return _connected;
   }

   // Return if MCIO is in error state
   public boolean getError() {
      return _error;
   }
   
   // Reset MCIO error
   public void resetError() {
      _error = false;
   }
   
   // Connect to a game server
   public int connect() {
      ConnectionFactory tempFactory;
      ConnectionDescriptor tempDescriptor;
      
      try {
         // Make connection to Minecraft game server
         tempFactory = new ConnectionFactory();
         //tempFactory.setPreferredTransportTypes(getAvailableTransports());
         tempDescriptor = tempFactory.getConnection("socket://" + _serverName);
         _serverStream = (StreamConnection)tempDescriptor.getConnection();
         _serverWriter = new DataOutputStream(_serverStream.openDataOutputStream());
         _serverReader = new DataInputStream(_serverStream.openDataInputStream());
      }
      catch (Exception e) {
         return RETURN_BADCONNECT;
      }
                 
      return RETURN_OK;
   }
   
   // Disconnect from a game server
   public int disconnect() {
      _connected = false;
      
      try {
         _serverWriter.close();
         _serverReader.close();
         _serverStream.close();
      }
      catch (Exception e) {
         return RETURN_BADDISCONNECT;
      }
      
      return RETURN_OK;
   }
   
   // Die from an error
   public void die(int passError) {
      Game.setStatus(MCIO.getErrorMessage(passError));
      disconnect();
      _connected = false; // If in an error state, we make sure connected is recorded as closed
      _error = true;
   }
      
   // Receive a Minecraft protocol packet.  The receiving workflow starts with this function, and then
   // implements the specific functionality using a "process" function (see below)
   public int receiveCommand() {
      byte packetCode;
      short tempLength;
      byte[] tempArray;

      try {     
         // Read in packet ID code
         packetCode = _serverReader.readByte();
         
         // Read in arguments
         for (int lcv = 0 ; lcv < ARGUMENTS_MAX ; lcv++) {
            // If we're done with arguments, exit
            if (_recvArgs[getByteIndex(packetCode)][lcv].argType == TYPE_NONE) break;

            // For each argument, lookup size and receive accordingly.  Special size is further processed
            switch (_recvArgs[getByteIndex(packetCode)][lcv].argType) {
               case TYPE_BOOLEAN: _recvArgs[getByteIndex(packetCode)][lcv].valBoolean = _serverReader.readBoolean(); break;
               case TYPE_BYTE: _recvArgs[getByteIndex(packetCode)][lcv].valByte = _serverReader.readByte(); break;
               case TYPE_SHORT: _recvArgs[getByteIndex(packetCode)][lcv].valShort = _serverReader.readShort(); break;
               case TYPE_INT: _recvArgs[getByteIndex(packetCode)][lcv].valInt = _serverReader.readInt(); break;
               case TYPE_LONG: _recvArgs[getByteIndex(packetCode)][lcv].valLong = _serverReader.readLong(); break;
               case TYPE_FLOAT: _recvArgs[getByteIndex(packetCode)][lcv].valFloat = _serverReader.readFloat(); break;
               case TYPE_DOUBLE: _recvArgs[getByteIndex(packetCode)][lcv].valDouble = _serverReader.readDouble(); break;
               case TYPE_STRING: tempLength = _serverReader.readShort();
                                 tempArray = new byte[tempLength * 2];
                                 _serverReader.readFully(tempArray, 0, tempArray.length);
                                 _recvArgs[getByteIndex(packetCode)][lcv].valString = new String(tempArray, 0, tempArray.length, "UTF-16BE"); 
                                 break;
               case TYPE_SPECIAL: receiveSpecial(packetCode, lcv); break;
            }
         }
      }
      catch (Exception e) {
         return RETURN_PACKETERROR;
      }

      // For commands we implement, call respective function here (mainly just logins/chat/time for now)
      switch (packetCode) {
         case ID_KEEP_ALIVE: return processKeepAlive(); 
         case ID_LOGIN_REQUEST: return processLoginRequest();
         case ID_HANDSHAKE: return processHandshake();
         case ID_CHAT_MESSAGE: return processChatMessage();
         case ID_TIME_UPDATE: return processTimeUpdate();
         case ID_SPAWN_POSITION: return processSpawnPosition();
         case ID_SPAWN_MOB: return processSpawnMob();
         case ID_SPAWN_PAINTING: return processSpawnPainting();
         case ID_PLAYER_ABILITIES: return processPlayerAbilities();
         case ID_DISCONNECT_KICK: return processDisconnectKick();
      }
      
      return RETURN_OK;
   }
   
   // Further receive special arguments which have dynamic sizes depending on other argument values
   public void receiveSpecial(byte passCode, int passParam) throws IOException {
      switch (passCode) {
         case ID_MAP_CHUNKS:
            // Array the size of argument 5
            _recvArgs[getByteIndex(passCode)][passParam].valArray = new byte[_recvArgs[getByteIndex(passCode)][5].valInt];
            _serverReader.readFully(_recvArgs[getByteIndex(passCode)][passParam].valArray, 0, _recvArgs[getByteIndex(passCode)][passParam].valArray.length);
            break;
         case ID_MULTI_BLOCK_CHANGE:
            // Array the size of argument 3
            _recvArgs[getByteIndex(passCode)][passParam].valArray = new byte[_recvArgs[getByteIndex(passCode)][3].valInt];
            _serverReader.readFully(_recvArgs[getByteIndex(passCode)][passParam].valArray, 0, _recvArgs[getByteIndex(passCode)][passParam].valArray.length);
            break; 
         case ID_EXPLOSION:
            // Array the size of argument 4 * 3
            _recvArgs[getByteIndex(passCode)][passParam].valArray = new byte[_recvArgs[getByteIndex(passCode)][4].valInt * 3];
            _serverReader.readFully(_recvArgs[getByteIndex(passCode)][passParam].valArray, 0, _recvArgs[getByteIndex(passCode)][passParam].valArray.length);
            break; 
         case ID_ITEM_DATA:
            // Array the size of argument 2
            _recvArgs[getByteIndex(passCode)][passParam].valArray = new byte[getByteIndex(_recvArgs[getByteIndex(passCode)][2].valByte)];
            _serverReader.readFully(_recvArgs[getByteIndex(passCode)][passParam].valArray, 0, _recvArgs[getByteIndex(passCode)][passParam].valArray.length);
            break;             
         case ID_PLUGIN_MESSAGE:
            // Array the size of argument 1
            _recvArgs[getByteIndex(passCode)][passParam].valArray = new byte[_recvArgs[getByteIndex(passCode)][1].valShort];
            _serverReader.readFully(_recvArgs[getByteIndex(passCode)][passParam].valArray, 0, _recvArgs[getByteIndex(passCode)][passParam].valArray.length);
            break;             
         case ID_PLAYER_BLOCK_PLACEMENT:
         case ID_CLICK_WINDOW:
         case ID_SET_SLOT:
         case ID_CREATIVE_INVENTORY_ACTION:
            // Receive single slot
            _recvArgs[getByteIndex(passCode)][passParam].valArray = receiveSlot(1);
            break;
         case ID_SET_WINDOW_ITEMS:
            // Receive argument 1 slots
            _recvArgs[getByteIndex(passCode)][passParam].valArray = receiveSlot(_recvArgs[getByteIndex(passCode)][1].valShort);
            break;
         case ID_SPAWN_MOB:
         case ID_ENTITY_METADATA:
            // Entity Metadata type
            _recvArgs[getByteIndex(passCode)][passParam].valArray = receiveEntityMeta();
            break;
         default:
            Game.addChat("Warning, unhandled special parameter: " + passCode + "\n");
            break;
      }
   }
   
   // Further receive slot type data (Player block placement, click window, set slot, etc)
   public byte[] receiveSlot(int passCount) throws IOException {
      Vector bufferVector;
      byte[] tempArray, tempArray2;
      int arraySize, arrayIndex;
      short tempID, tempMeta, tempLength;
      byte tempCount;
            
      bufferVector = new Vector();
      for (int lcv = 0 ; lcv < passCount ; lcv++) {
         tempID = _serverReader.readShort();
         
         if (tempID == -1) {
            // Empty Slot
            bufferVector.addElement(ByteBuffer.allocateDirect(2));
            ((ByteBuffer)bufferVector.lastElement()).putShort(tempID);
         }
         else if (enchantFlag(tempID)) {
            // Enchantable item
            tempCount = _serverReader.readByte();
            tempMeta = _serverReader.readShort();
            tempLength = _serverReader.readShort();
            tempArray2 = new byte[0];
            
            if (tempLength > -1) {
               bufferVector.addElement(ByteBuffer.allocateDirect(7 + tempLength));
               tempArray2 = new byte[tempLength];
               _serverReader.readFully(tempArray2, 0, tempLength);
            }
            else {
               bufferVector.addElement(ByteBuffer.allocateDirect(7));
            }
            
            ((ByteBuffer)bufferVector.lastElement()).putShort(tempID);
            ((ByteBuffer)bufferVector.lastElement()).put(tempCount);
            ((ByteBuffer)bufferVector.lastElement()).putShort(tempMeta);               
            ((ByteBuffer)bufferVector.lastElement()).putShort(tempLength);
            if (tempLength > -1) ((ByteBuffer)bufferVector.lastElement()).put(tempArray2);
         }
         else {
            // Normal item
            bufferVector.addElement(ByteBuffer.allocateDirect(5));
            ((ByteBuffer)bufferVector.lastElement()).putShort(tempID);
            ((ByteBuffer)bufferVector.lastElement()).put(_serverReader.readByte());
            ((ByteBuffer)bufferVector.lastElement()).putShort(_serverReader.readShort());               
         }
      }

      // Copy all bytebuffers over to array
      arraySize = 0;
      for (int lcv = 0 ; lcv < bufferVector.size(); lcv++) {
         arraySize += ((ByteBuffer)bufferVector.elementAt(lcv)).capacity();
      }
      
      tempArray = new byte[arraySize];
   
      arrayIndex = 0;
      for (int lcv = 0 ; lcv < bufferVector.size(); lcv++) {
         ((ByteBuffer)bufferVector.elementAt(lcv)).rewind();
         ((ByteBuffer)bufferVector.elementAt(lcv)).get(tempArray, arrayIndex, ((ByteBuffer)bufferVector.elementAt(lcv)).capacity());
         arrayIndex = ((ByteBuffer)bufferVector.elementAt(lcv)).capacity();
      }     
      
      return tempArray;
   }
   
   // Further receive entity metadata (spawn mob and entity metadata) 
   public byte[] receiveEntityMeta() throws IOException {
      Vector bufferVector;
      byte[] tempArray;
      short stringSize;
      int arraySize, arrayIndex;
      byte tempByte;
      
      bufferVector = new Vector();
      while ((tempByte = _serverReader.readByte()) != 127) {
         switch ((tempByte >> 5) & 0x07) {
            case 0:
               bufferVector.addElement(ByteBuffer.allocateDirect(2));
               ((ByteBuffer)bufferVector.lastElement()).put(tempByte);
               ((ByteBuffer)bufferVector.lastElement()).put(_serverReader.readByte());
               break;
            case 1: 
               bufferVector.addElement(ByteBuffer.allocateDirect(3));
               ((ByteBuffer)bufferVector.lastElement()).put(tempByte);
               ((ByteBuffer)bufferVector.lastElement()).putShort(_serverReader.readShort());
               break;
            case 2: 
               bufferVector.addElement(ByteBuffer.allocateDirect(5));
               ((ByteBuffer)bufferVector.lastElement()).put(tempByte);
               ((ByteBuffer)bufferVector.lastElement()).putInt(_serverReader.readInt());
               break;
            case 3: 
               bufferVector.addElement(ByteBuffer.allocateDirect(5));
               ((ByteBuffer)bufferVector.lastElement()).put(tempByte);
               ((ByteBuffer)bufferVector.lastElement()).putFloat(_serverReader.readFloat());
               break;
            case 4: 
               stringSize = _serverReader.readShort();
               bufferVector.addElement(ByteBuffer.allocateDirect(3+stringSize));
               ((ByteBuffer)bufferVector.lastElement()).put(tempByte);
               ((ByteBuffer)bufferVector.lastElement()).putShort(stringSize);
               for (int lcv = 0 ; lcv < stringSize ; lcv++) {
                  ((ByteBuffer)bufferVector.lastElement()).put(_serverReader.readByte());
               }
               break;
            case 5: 
               bufferVector.addElement(ByteBuffer.allocateDirect(6));
               ((ByteBuffer)bufferVector.lastElement()).put(tempByte);
               ((ByteBuffer)bufferVector.lastElement()).putShort(_serverReader.readShort());
               ((ByteBuffer)bufferVector.lastElement()).put(_serverReader.readByte());
               ((ByteBuffer)bufferVector.lastElement()).putShort(_serverReader.readShort());
               break;
            case 6: 
               bufferVector.addElement(ByteBuffer.allocateDirect(13));
               ((ByteBuffer)bufferVector.lastElement()).put(tempByte);
               ((ByteBuffer)bufferVector.lastElement()).putInt(_serverReader.readInt());
               ((ByteBuffer)bufferVector.lastElement()).putInt(_serverReader.readInt());
               ((ByteBuffer)bufferVector.lastElement()).putInt(_serverReader.readInt());
               break;
            default:
               Game.addChat("Warning, unhandled meta type: " + (tempByte >> 5) + "\n");
               break;
         }
      }
      
      // Copy all bytebuffers over to array
      arraySize = 0;
      for (int lcv = 0 ; lcv < bufferVector.size(); lcv++) {
         arraySize += ((ByteBuffer)bufferVector.elementAt(lcv)).capacity();
      }
      
      tempArray = new byte[arraySize];
   
      arrayIndex = 0;
      for (int lcv = 0 ; lcv < bufferVector.size(); lcv++) {
         ((ByteBuffer)bufferVector.elementAt(lcv)).rewind();
         ((ByteBuffer)bufferVector.elementAt(lcv)).get(tempArray, arrayIndex, ((ByteBuffer)bufferVector.elementAt(lcv)).capacity());
         arrayIndex = ((ByteBuffer)bufferVector.elementAt(lcv)).capacity();
      }

      return tempArray;
   }
   
   // Send a Minecraft protocol packet.  The sending workflow starts with a send function to load
   // the arguments into the send Array, and then calls this function (opposite workflow from receiveCommand)
   public int sendCommand(byte passCode) {
      try {
         // Send packet ID code
         _serverWriter.writeByte(passCode);
         // Send arguments
         for (int lcv = 0 ; lcv < ARGUMENTS_MAX ; lcv++) {
            // If we're done with arguments, exit
            if (_sendArgs[getByteIndex(passCode)][lcv].argType == TYPE_NONE) break;
                       
            switch (_sendArgs[getByteIndex(passCode)][lcv].argType) {
               case TYPE_BOOLEAN: _serverWriter.writeBoolean(_sendArgs[getByteIndex(passCode)][lcv].valBoolean); break;
               case TYPE_BYTE: _serverWriter.writeByte(_sendArgs[getByteIndex(passCode)][lcv].valByte); break;
               case TYPE_SHORT: _serverWriter.writeShort(_sendArgs[getByteIndex(passCode)][lcv].valShort); break;
               case TYPE_INT: _serverWriter.writeInt(_sendArgs[getByteIndex(passCode)][lcv].valInt); break;
               case TYPE_LONG: _serverWriter.writeLong(_sendArgs[getByteIndex(passCode)][lcv].valLong); break;
               case TYPE_FLOAT: _serverWriter.writeFloat(_sendArgs[getByteIndex(passCode)][lcv].valFloat); break;
               case TYPE_DOUBLE: _serverWriter.writeDouble(_sendArgs[getByteIndex(passCode)][lcv].valDouble); break;
               case TYPE_STRING: _serverWriter.writeShort(_sendArgs[getByteIndex(passCode)][lcv].valString.length());
                                 _serverWriter.write(_sendArgs[getByteIndex(passCode)][lcv].valString.getBytes("UTF-16BE")); 
                                 break;
            }
         }
      }
      catch (IOException e) {
         return RETURN_PACKETERROR;
      }
      
      return RETURN_OK;
   }
   
   public int sendKeepAlive(int passValue) {
      _sendArgs[getByteIndex(ID_KEEP_ALIVE)][0].valInt = passValue;
      
      return sendCommand(ID_KEEP_ALIVE);      
   }
   
   public int processKeepAlive() {
      return sendKeepAlive(_recvArgs[getByteIndex(ID_KEEP_ALIVE)][0].valInt);
   }

   public int sendLoginRequest() {
      _sendArgs[getByteIndex(ID_LOGIN_REQUEST)][0].valInt = Integer.parseInt(_protocolVer);
      _sendArgs[getByteIndex(ID_LOGIN_REQUEST)][1].valString = _userName;
      
      return sendCommand(ID_LOGIN_REQUEST);
   }
   
   public int processLoginRequest() {
      _connected = true;
      
      return RETURN_OK;
   }   
   
   public int sendHandshake() {
      _sendArgs[getByteIndex(ID_HANDSHAKE)][0].valString = _userName + ";" + _serverName;
      
      return sendCommand(ID_HANDSHAKE);
   }
   
   public int processHandshake() {
      _serverHash = _recvArgs[getByteIndex(ID_HANDSHAKE)][0].valString;
      
      // Join Minecraft.net server if gameserver in online mode
      if (_serverHash.equalsIgnoreCase("-") == false) {
         if (requestSession() == false) return RETURN_BADAUTH;
         if (requestJoin() == false) return RETURN_BADJOIN;
      }

      return RETURN_OK;
   }
  
   public int sendChatMessage(String passText) {
      if (passText.length() > 100) passText = passText.substring(0, 100);
      _sendArgs[getByteIndex(ID_CHAT_MESSAGE)][0].valString = passText;
      
      return sendCommand(ID_CHAT_MESSAGE);
   }
   
   public int processChatMessage() {
      Game.addChat(_recvArgs[getByteIndex(ID_CHAT_MESSAGE)][0].valString + "\n");
      
      return RETURN_OK;
   }
   
   public int processTimeUpdate() {
      Game.setTime(_recvArgs[getByteIndex(ID_TIME_UPDATE)][0].valLong);
      
      return RETURN_OK;
   }

   // Current does nothing - here as example   
   public int processSpawnPosition() {
      return RETURN_OK;
   }
   
   // Current does nothing - here as example
   public int processSpawnMob() {
      //ChatScreen.addOutput("Spawning Mob (ID, Type): " + _recvArgs[getByteIndex(ID_SPAWN_MOB)][0].valInt + "," + _recvArgs[getByteIndex(ID_SPAWN_MOB)][1].valByte, true);
      
      return RETURN_OK;
   }

   // Current does nothing - here as example   
   public int processSpawnPainting() {
      //ChatScreen.addOutput("Spawning Painting (ID, Name): " + _recvArgs[getByteIndex(ID_SPAWN_PAINTING)][0].valInt + "," + _recvArgs[getByteIndex(ID_SPAWN_PAINTING)][1].valString, true);
      
      return RETURN_OK;
   }   
   
   // Current does nothing - here as example   
   public int processPlayerAbilities() {      
      return RETURN_OK;
   }

   public int sendDisconnectKick(String passMessage) {
      _sendArgs[getByteIndex(ID_DISCONNECT_KICK)][0].valString = passMessage;
      
       return sendCommand(ID_DISCONNECT_KICK);
   }
   
   public int processDisconnectKick() {
      Game.addChat(_recvArgs[getByteIndex(ID_DISCONNECT_KICK)][0].valString + "\n");
      
      return disconnect();
   }
}

